#define _XOPEN_SOURCE

/* Includes */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <unistd.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <sys/time.h>
#include <dirent.h>
#include <signal.h>
#include <math.h>
#include <gtk/gtk.h>
#include <pthread.h>
#include <semaphore.h>
#include <fcntl.h>
#include <mqueue.h>
#include "vorbis/codec.h"
#include "vorbis/vorbisfile.h"
#include "id3.h"
#include "FLAC/format.h"
#include "FLAC/metadata.h"
#include "audconvert.h"
#include "gui.h"
#include "drivers.h"
/* End Includes */

/* getopt stuff */
extern int      optopt;
extern int      optind;
extern char    *optarg;
extern int      opterr;

int             getopt (int, char *const *, const char *);
/* end getopt stuff */

/* non-ansi defines */
int snprintf(char *str, size_t size, const char *format, ...);

/* private function defines */
void *process_file( void *shared_file_info );
void *file_write_thread( void *ptr );

/* Global thread variables. */
static int threads = 1;                 /* Number of threads to run. */
static int processes[MAX_THREADS];      /* Array to keep track of which thread
                                         * we're using. */
static pid_t children[MAX_THREADS * 3]; /* Array of children PIDs.
                                         * Used for shutting down cleanly. */
/* Mutex to make sure we get the file_info data before it's removed. */
static pthread_mutex_t mutex1 = PTHREAD_MUTEX_INITIALIZER;

/* Mutex to make sure only one directory is created at a time. */
static pthread_mutex_t dirmutex = PTHREAD_MUTEX_INITIALIZER;

static int files_processed = 0;         /* Count of the number of files we've
                                         * processed. */

/* User Setable Vars */
static char input_dir_str[MAXLENGTH] = "";
static char output_dir_str[MAXLENGTH] = "";

static int flatten = 0;

static int justgo = 0;
static int benchmark = 0;

static int oggin = 1;
static int mp3in = 1;
static int flacin = 1;

static int lameout = 1;
static int oggout = 0;

static int output_bitrate = 192;

/* Print the help message. */
int help_msg( void ) {
    printf( "\nUsage: audconvert [options]\n\n"
            "Options:\n"
            "--------\n\n"
            "-b\t\tBenchmark.  Start process on execute, close on completion,\n"
            "\t\tand print statistics.\n"
            "-f\t\tFlatten the directory structure. All output files in one\n"
            "\t\tdirectory.\n"
            "-g\t\tGo. Don't wait for a button press.\n"
            "-h\t\tDisplay this help text and exit.\n"
            "-i directory\tSpecify the directory of the input files.\n"
            "-o directory\tSpecify the directory of the output files.\n"
            "-t threads\tMaximum number of threads to use for conversion.\n"
            "\t\t(Maximum value: %d)\n"
            "-v\t\tDisplay version number.\n"
            "\n", MAX_THREADS );

    return SUCCESS;
}

/* Set the input or output directory from the GUI. */
int set_input_ouput_dir( const char *inputdir, const char *outputdir )
{
    strncpy( input_dir_str, inputdir, MAXLENGTH );
    strncpy( output_dir_str, outputdir, MAXLENGTH );

    return SUCCESS;
}

/* Set a value of a var from the GUI. */
int set_value( int field, int value )
{
    switch (field)
    {
        case FLATTEN:
            flatten = value;
            break;
#if 0 /* Stubbed for completeness.  Don't use this. */
        case JUSTGO:
            justgo = value;
            break;
#endif
#if 0 /* Stubbed for completeness.  Don't use this. */
        case BENCHMARK:
            benchmark = value;
            break;
#endif
        case OUTPUTBITRATE:
            output_bitrate = value;
            break;
        case OGGIN:
            oggin = value;
            break;
        case MP3IN:
            mp3in = value;
            break;
        case FLACIN:
            flacin = value;
            break;
        case LAMEOUT:
            lameout = value;
            break;
        case OGGOUT:
            oggout = value;
            break;
        case THREADS:
            threads = value;
            break;
        default:
            return FAILURE;
    }

    return SUCCESS;
}

/* Get a value for GUI setup. */
int get_value( int field )
{
    switch (field)
    {
        case FLATTEN:
            return flatten;
        case JUSTGO:
            return justgo;
        case BENCHMARK:
            return benchmark;
        case OUTPUTBITRATE:
            return output_bitrate;
        case OGGIN:
            return oggin;
        case MP3IN:
            return mp3in;
        case FLACIN:
            return flacin;
        case LAMEOUT:
            return lameout;
        case OGGOUT:
            return oggout;
        case THREADS:
            return threads;
        default:
            return -1;
    }

    return SUCCESS;
}

/* All all the child threads started with fork(). */
void kill_children()
{
    int i = 0;
    
    for ( i = 0; i < (MAX_THREADS * 3); i++ )
    {
        if ( children[i] != -1 )
        {
            if ( kill( children[i], SIGTERM ) != 0 )
            {
                perror( "kill() failed" );
            }
        }
    }
}

/* Recurse into the directory and look for supported input files. */
int enter_dir( char *dir_name, char *output_dir, sem_t *thread_count )
{
    DIR *input_dir = NULL;                  /* Input directory pointer */
    struct dirent *input_dir_list = NULL;   /* Directory listing */
    struct stat stat_file_info;             /* Directory/File Info */
    char *full_path = NULL;                 /* Full path of directory/file */
    char *output_dir_new = NULL;            /* The directory where we'll 
                                               output new files. */
    int i = 0;                              /* Temp integer. */

    char log_message[MAXLENGTH];

    pthread_t threadid;
    struct encode_data file_info;
    
    struct bar_info_msg bar_data;
    mqd_t mque;

    file_info.inputformat = -1;

    /* Open the directory that we were passed. */
    if ( ( input_dir = opendir( dir_name ) ) == NULL )
    {
        snprintf( log_message, MAXLENGTH, "Input directory open failed: %s\n",
                  dir_name );
        print_log( log_message ); 
        return FAILURE;
    }

    mque = mq_open( AUDCONVERT_MQUE, O_WRONLY, NULL );
    if ( mque == -1 )
    {
        snprintf( log_message, MAXLENGTH,
                  "mq_open() failure: %s\n", strerror( errno ) );
        print_log( log_message ); 
    }

    /* Read the contents of the current directory. */
    while ( ( input_dir_list = readdir( input_dir ) ) != NULL )
    {
        if ( ( strcmp( ".", input_dir_list->d_name ) == 0 ) ||
             ( strcmp( "..", input_dir_list->d_name ) == 0 ) )
        {
            /* Ignore . and .. directories. */
        } else {
            /* Create the new full path. */
            full_path = malloc( ( strlen( dir_name ) + 
                                  strlen( input_dir_list->d_name ) + 2 ) *
                                sizeof( char ) );
            strcpy( full_path, dir_name );
            if ( full_path[ strlen( full_path ) - 1 ] != '/' )
            {
                strcat( full_path, "/" );
            }
            strcat( full_path, input_dir_list->d_name );

            /* Get the directory/file info. */
            stat( full_path, &stat_file_info );

            /* Is this a directory? */
            if ( S_ISDIR( stat_file_info.st_mode ) )
            {
#ifdef DEBUG
                printf( "DIRECTORY\n" );
#endif
                /* Create the output path. */
                output_dir_new = malloc( ( strlen( output_dir ) +
                                           strlen( input_dir_list->d_name ) +
                                           2 ) * sizeof( char ) );
                strcpy( output_dir_new, output_dir );
                if ( output_dir_new[ strlen( output_dir_new ) - 1 ] != '/' )
                {
                    strcat( output_dir_new, "/" );
                }
                strcat( output_dir_new, input_dir_list->d_name );

                /* Recurse into the new directory. */
                enter_dir( full_path, output_dir_new, thread_count );

                /* Cleanup memory. */
                free( output_dir_new );
            } else if ( S_ISREG( stat_file_info.st_mode ) ) {
                /* This is a file, not a directory. */
                /* Is the an Ogg Vorbis file? */
                if ( oggin && (strstr( full_path, ".ogg" ) != NULL) )
                {
#ifdef DEBUG
                    fprintf( stderr, "Found ogg file: %s\n", full_path );
#endif

                    file_info.inputformat = OGG;
                }

                if ( mp3in && (strstr( full_path, ".mp3" ) != NULL) )
                {
#ifdef DEBUG
                    fprintf( stderr, "Found mp3 file: %s\n", full_path );
#endif

                    file_info.inputformat = MP3;
                }

                if ( flacin && (strstr( full_path, ".flac" ) != NULL) )
                {
#ifdef DEBUG
                    fprintf( stderr, "Found flac file: %s\n", full_path );
#endif

                    file_info.inputformat = FLAC;
                }
                
                if ( file_info.inputformat != -1 )
                {
                    /* Wait for a thread to become available. */
#ifdef DEBUG
                    fprintf( stderr, "Waiting on available thread.\n" );
#endif
                    sem_wait( thread_count );

                    for ( i = 0; i < threads; i++ )
                    {
                        if ( processes[i] == -1 )
                        {
                            processes[i] = 0;
                            break;
                        }
                    }

#ifdef DEBUG
                    fprintf( stderr, "Got available thread: %d.\n", i );
#endif
                    
                    /* Fork for encoding. */
                    bar_data.mtype = SBARUPDATE;
                    bar_data.bar_id = i;
                    snprintf( bar_data.bar_value, MAXLENGTH, "..%s", 
                              full_path+strlen(input_dir_str) );
                    
                    if ( mq_send( mque, (char *) &bar_data,
                                  sizeof( bar_data ), 0 ) != 0 )
                    {
                        snprintf( log_message, MAXLENGTH,
                                  "mq_send() failed: %s\n", strerror( errno ) );
                        print_log( log_message ); 
                    } else {
#ifdef DEBUG
                        printf( "SEND: sbar[%d] = %s\n", i,
                                bar_data.bar_value );
#endif
                    }

                    strcpy( file_info.file_name, full_path );
                    strcpy( file_info.output_dir, output_dir );
                    strcpy( file_info.output_filename, input_dir_list->d_name );
                    file_info.thread_num = i;

#ifdef DEBUG
                    fprintf( stderr, "Creating thread..." );
#endif
                    pthread_mutex_lock( &mutex1 );
                    pthread_create( &threadid, NULL, process_file, &file_info );
                    pthread_detach( threadid );
#ifdef DEBUG
                    fprintf( stderr, "Done.\n" );
#endif
                    files_processed++;
                } else {
                    /* If not an Ogg Vorbis file, ignore. */
#ifdef DEBUG
                    printf( "REGULAR FILE\n" );
#endif
                }
            } else {
                /* If it's not a directory or an Ogg Vorbis file, ignore. */
            }

            pthread_mutex_lock( &mutex1 );
            /* Cleanup memory. */
            free( full_path );
            pthread_mutex_unlock( &mutex1 );
        }
    }

    if ( mq_close( mque ) != 0 )
    {
        snprintf( log_message, MAXLENGTH,
                  "mq_close() failed: %s\n", strerror( errno ) );
        print_log( log_message ); 
    }

    /* Close directory. */
    closedir( input_dir );

    return SUCCESS;
}

/* Read the provided string and parse it into the file_info struct. */
int parse_vorbis_comment( char *string, struct encode_data *file_info )
{
    /* Get the title. */
    if ( strncasecmp( string, "title", 5 ) == 0 )
    {
        strncpy( file_info->title, &string[ 6 ], MAXLENGTH );
#ifdef DEBUG
        printf( "TITLE = %s\n", file_info->title );
#endif
        return 1;
    }

    /* Get the artist. */
    if ( strncasecmp( string, "artist", 6 ) == 0 )
    {
        strncpy( file_info->artist, &string[ 7 ], MAXLENGTH );
#ifdef DEBUG
        printf( "ARTIST = %s\n", file_info->artist );
#endif
        return 1;
    }
                
    /* Get the genre. */
    if ( strncasecmp( string, "genre", 5 ) == 0 )
    {
        strncpy( file_info->genre, &string[ 6 ], MAXLENGTH );
#ifdef DEBUG
        printf( "GENRE = %s\n", file_info->genre );
#endif
        return 1;
    }

    /* Get the album. */
    if ( strncasecmp( string, "album", 5 ) == 0 )
    {
        strncpy( file_info->album, &string[ 6 ], MAXLENGTH );
#ifdef DEBUG
        printf( "ALBUM = %s\n", file_info->album );
#endif
        return 1;
    }

    /* Get the tracknumber. */
    if ( strncasecmp( string, "tracknumber", 11 ) == 0 )
    {
        strncpy( file_info->tracknumber, &string[ 12 ], MAXLENGTH );
#ifdef DEBUG
        printf( "TRACKNUMBER = %s\n", file_info->tracknumber );
#endif
        return 1;
    }
 
    /* Get the date. */
    if ( strncasecmp( string, "date", 4 ) == 0 )
    {
        strncpy( file_info->date, &string[ 5 ], MAXLENGTH );
#ifdef DEBUG
        printf( "DATE = %s\n", file_info->date );
#endif
        return 1;
    }
 
    /* Get the comment. */
    if ( strncasecmp( string, "comment", 7 ) == 0 )
    {
        strncpy( file_info->comment, &string[ 8 ], MAXLENGTH );
#ifdef DEBUG
        printf( "COMMENT = %s\n", file_info->comment );
#endif
        return 1;
    }

    return 0;
}

/* Create directories and subdirectories. */
int create_dir( const char *dir_name, mode_t mode )
{
    char log_message[MAXLENGTH];    /* Hold log messages. */
    char lower_dir[MAXLENGTH];      /* Lower directory to check. */
    struct stat dir_info;           /* Info of output directory. */
    int length = 0, i = 0;          /* For finding sub-directories. */
    int status = 0;                 /* General status. */

    /* Check the status of the directory passed. */
    status = stat( dir_name, &dir_info );
    if ( status == -1 )
    {
        /* If the directory doesn't exist, make sure the sub directory does. */
        if ( errno == ENOENT )
        {
            /* Get sub directory. */
            length = strlen( dir_name );
            for ( i = 1; i < length; i++ )
            {
                if ( dir_name[ length - i ] == '/' )
                {
                    break;
                }
            }

            strncpy( lower_dir, dir_name, length - i );
            lower_dir[ length - i + 1 ] = '\0';
            
            /* Create the sub directory. */
            status = create_dir( lower_dir, mode );
            if ( status == FAILURE )
            {
                return FAILURE;
            }
        } else {
            snprintf( log_message, MAXLENGTH, "ERROR: stat(): %s\n",
                      strerror( errno ) );
            print_log( log_message );

            return FAILURE;
        }

        /* Create the directory requested now. */
        status = mkdir( dir_name, 0755 );
        if ( status == -1 )
        {
            if ( errno != EEXIST )
            {
                snprintf( log_message, MAXLENGTH, "ERROR: mkdir() %s: %s\n",
                          dir_name, strerror( errno ) );
                print_log( log_message );

                return FAILURE;
            }
        }
    } else {
        /* If the stat() suceeded, make sure the entry is a directory. */
        if ( !S_ISDIR( dir_info.st_mode ) )
        {
            snprintf( log_message, MAXLENGTH,
                      "ERROR: output directory is not a directory.\n" );
            print_log( log_message );

            return FAILURE;
        }
    }

    return SUCCESS;
}

/* Process the file. */
void *process_file( void *shared_file_info )
{
    struct encode_data file_info;

    char log_message[MAXLENGTH];
    
    OggVorbis_File vf;          /* Ogg Vorbis file handle. */
    char **ptr;
    vorbis_info *vi;

    FLAC__uint32 comments = 0;
    FLAC__StreamMetadata *tags;

    ID3Tag *id3info;
    ID3Frame *id3frame;
    ID3Field *id3field;
    char tmpstr[MAXLENGTH];

    char *complete_output = NULL;   /* Complete path of output file. */
    
    int i = 0, j = 0;           /* Temp vars. */

    int return_status = SUCCESS;
    
    FILE *input_file = NULL;    /* Input file handle. */

    sem_t *thread_count;

    int status = 0;             /* General status. */

#if 0
    pthread_t write_threadid;
    pthread_t input_threadid;
    pthread_t output_threadid;
#else
    pid_t child1;
    pid_t child2;
    pid_t child3;
#endif

    memcpy( &file_info, shared_file_info, sizeof( struct encode_data ) );
    pthread_mutex_unlock( &mutex1 );

    thread_count = sem_open( AUDCONVERT_SEM, 0 );
    if ( thread_count == SEM_FAILED )
    {
        snprintf( log_message, MAXLENGTH,
                  "semaphore create failed: %s\n", strerror( errno ) );
        print_log( log_message ); 
    }
 
    /* Fill in file info based on format. */
    switch ( file_info.inputformat )
    {
        case OGG:
#ifdef DEBUG
            fprintf( stderr, "Processing OGG file...\n" );
#endif
            /* Open the input file. */
            if ( ( input_file = fopen( file_info.file_name, "rb" ) ) == NULL )
            {
                snprintf( log_message, MAXLENGTH,
                          "Opening of file \"%s\" failed!\n",
                          file_info.file_name );
                print_log( log_message ); 

                sem_post( thread_count );
                return_status = FAILURE;
                pthread_exit( &return_status );
            }
        
            /* Open the file as an Ogg Vorbis file. */
            if ( ov_open( input_file, &vf, NULL, 0 ) < 0 )
            {
                snprintf( log_message, MAXLENGTH,
                          "Input does not appear to be an Ogg bitstream.\n" );
                print_log( log_message ); 

                sem_post( thread_count );
                return_status = FAILURE;
                pthread_exit( &return_status );
            }
    
            /* Pull the comment info from the Ogg Vorbis file. */
            ptr = ov_comment( &vf, -1 )->user_comments;

            /* Get the Ogg Vorbis file info. */
            vi = ov_info( &vf, -1 );

            /* Sort through the comment field. */
            for ( j = 0; j < ov_comment( &vf, -1 )->comments; j++ )
            {
                if ( parse_vorbis_comment( ptr[ j ], &file_info ) != 1 )
                {
                    snprintf( log_message, MAXLENGTH,
                              "Comment string \"%s\" not recognized.",
                              ptr[ j ] );
                    print_log( log_message ); 
                }
            }

#if 0
            fprintf( stderr, "\nBitstream is %d channel, %ldHz\n",
                     vi->channels, vi->rate );
            fprintf( stderr,
                     "\nUpper Rate: %ld\nNominal Rate: %ld\nLower Rate: %ld\n",
                     vi->bitrate_upper, vi->bitrate_nominal, vi->bitrate_lower );
            fprintf( stderr, "\nNominal Rate: %ld\n", vi->bitrate_nominal );
            fprintf( stderr, "\nDecoded length: %ld samples\n",
                     (long) ov_pcm_total( &vf, -1 ) );
            fprintf( stderr, "Encoded by: %s\n\n", ov_comment( &vf, -1)->vendor );
#endif

            ov_clear( &vf );
            break;
        case MP3:
#ifdef DEBUG
            fprintf( stderr, "Processing MP3 file...\n" );
#endif
            id3info = ID3Tag_New();
            ID3Tag_Link( id3info, file_info.file_name );
            
            id3frame = ID3Tag_FindFrameWithID( id3info, ID3FID_TITLE );
            id3field = ID3Frame_GetField( id3frame, ID3FN_TEXT );
            ID3Field_GetASCII( id3field, file_info.title, MAXLENGTH );
            
            id3frame = ID3Tag_FindFrameWithID( id3info, ID3FID_LEADARTIST );
            id3field = ID3Frame_GetField( id3frame, ID3FN_TEXT );
            ID3Field_GetASCII( id3field, file_info.artist, MAXLENGTH );
            
            id3frame = ID3Tag_FindFrameWithID( id3info, ID3FID_CONTENTTYPE );
            id3field = ID3Frame_GetField( id3frame, ID3FN_TEXT );
            ID3Field_GetASCII( id3field, tmpstr, MAXLENGTH );
            i = 0;
            for ( j = 0; j < strlen( tmpstr ); j++ )
            {
                if ( (tmpstr[j] != '(') && (tmpstr[j] != ')') )
                {
                    file_info.genre[i++] = tmpstr[j];
                }
            }
            file_info.genre[i] = '\0';
            strcpy( tmpstr, file_info.genre );
            if ( atoi(tmpstr) < ID3_NR_OF_V1_GENRES && atoi(tmpstr) >= 0 )
            {
                strcpy( file_info.genre,
                        ID3_v1_genre_description[atoi(tmpstr)] );
            }
            
            id3frame = ID3Tag_FindFrameWithID( id3info, ID3FID_ALBUM );
            id3field = ID3Frame_GetField( id3frame, ID3FN_TEXT );
            ID3Field_GetASCII( id3field, file_info.album, MAXLENGTH );
            
            id3frame = ID3Tag_FindFrameWithID( id3info, ID3FID_TRACKNUM );
            id3field = ID3Frame_GetField( id3frame, ID3FN_TEXT );
            ID3Field_GetASCII( id3field, file_info.tracknumber, MAXLENGTH );
            
            id3frame = ID3Tag_FindFrameWithID( id3info, ID3FID_YEAR );
            id3field = ID3Frame_GetField( id3frame, ID3FN_TEXT );
            ID3Field_GetASCII( id3field, file_info.date , MAXLENGTH );

            id3frame = ID3Tag_FindFrameWithID( id3info, ID3FID_COMMENT );
            id3field = ID3Frame_GetField( id3frame, ID3FN_TEXT );
            ID3Field_GetASCII( id3field, file_info.comment, MAXLENGTH );

            break;
        case FLAC:
#ifdef DEBUG
            fprintf( stderr, "Processing FLAC file...\n" );
#endif
            if ( FLAC__metadata_get_tags( file_info.file_name, &tags) != true )
            {
                snprintf( log_message, MAXLENGTH,
                          "Error reading FLAC file.\n" );
                print_log( log_message );
            }

            comments = tags->data.vorbis_comment.num_comments;
            for ( i = 0; i < comments; i++ )
            {
                if ( parse_vorbis_comment(
                            (char *)tags->data.vorbis_comment.comments[i].entry,
                            &file_info ) != 1 )
                {
                    snprintf( log_message, MAXLENGTH,
                              "Comment string \"%s\" not recognized.",
                              tags->data.vorbis_comment.comments[i].entry );
                    print_log( log_message ); 
                }
            }

            FLAC__metadata_object_delete( tags );

            break;
        default:
            break;
    }

    /* Put together the complete output filename. */
    if ( flatten == 0 )
    {
        complete_output = malloc( ( strlen( file_info.output_dir ) + 
                                 strlen( file_info.output_filename ) + 2 ) *
                                 sizeof( char ) );
        strcpy( complete_output, file_info.output_dir );
    } else {
        complete_output = malloc( ( strlen( output_dir_str ) + 
                                 strlen( file_info.output_filename ) + 2 ) *
                                 sizeof( char ) );
        strcpy( complete_output, output_dir_str );
    }
    strcat( complete_output, "/" );
    strcat( complete_output, file_info.output_filename );
    if ( lameout )
    {
        if ( file_info.inputformat == FLAC )
        {
            complete_output[ strlen( complete_output ) - 4 ] = 'm';
            complete_output[ strlen( complete_output ) - 3 ] = 'p';
            complete_output[ strlen( complete_output ) - 2 ] = '3';
            complete_output[ strlen( complete_output ) - 1 ] = '\0';
        } else {
            complete_output[ strlen( complete_output ) - 3 ] = 'm';
            complete_output[ strlen( complete_output ) - 2 ] = 'p';
            complete_output[ strlen( complete_output ) - 1 ] = '3';
        }
    }

    if ( oggout )
    {
        if ( file_info.inputformat == FLAC )
        {
            complete_output[ strlen( complete_output ) - 4 ] = 'o';
            complete_output[ strlen( complete_output ) - 3 ] = 'g';
            complete_output[ strlen( complete_output ) - 2 ] = 'g';
            complete_output[ strlen( complete_output ) - 1 ] = '\0';
        } else {
            complete_output[ strlen( complete_output ) - 3 ] = 'o';
            complete_output[ strlen( complete_output ) - 2 ] = 'g';
            complete_output[ strlen( complete_output ) - 1 ] = 'g';
        }
    }

    if ( strlen( complete_output ) >= MAXLENGTH )
    {
        snprintf( log_message, MAXLENGTH,
                  "ERROR: OUTPUT FILENAME TOO BIG!\n" );
        print_log( log_message ); 
    } else {
        strncpy( file_info.complete_output, complete_output, MAXLENGTH );
    }
        
    pthread_mutex_lock( &dirmutex );
    if ( flatten == 0 )
    {
        status = create_dir( file_info.output_dir, 0755 );
        if ( status == FAILURE )
        {
            pthread_mutex_unlock( &dirmutex );
            sem_post( thread_count );
            return_status = FAILURE;
            pthread_exit( &return_status );
        }
    } else {
        status = create_dir( output_dir_str, 0755 );
        if ( status == FAILURE )
        {
            pthread_mutex_unlock( &dirmutex );
            sem_post( thread_count );
            return_status = FAILURE;
            pthread_exit( &return_status );
        }
    }
    pthread_mutex_unlock( &dirmutex );
    
#if 0
    /* WRITE THREAD */
    pthread_create( &write_threadid, NULL, file_write_thread, &file_info );
    sleep( 1 );

    /* DECODE THREAD */
    pthread_create( &input_threadid, NULL, ogg_input_thread, &file_info );
    sleep( 1 );

    /* ENCODE THREAD */
    pthread_create( &output_threadid, NULL, lame_output_thread, &file_info );
    sleep( 1 );

    /* pthread_join( output_threadid, NULL ); */
    sleep( 60 );
#else

    child1 = fork();
    if ( child1 == 0 )
    {
        file_write_thread( &file_info );

        exit(0);
    } else {
        children[(file_info.thread_num*3)] = child1;
    }

    child2 = fork();
    if ( child2 == 0 )
    {
        switch ( file_info.inputformat )
        {
            case OGG:
                ogg_input_thread( &file_info );
                break;
            case MP3:
                mpg123_input_thread( &file_info );
                break;
            case FLAC:
                flac_input_thread( &file_info );
                break;
            default:
                break;
        }

        exit(0);
    } else {
        children[(file_info.thread_num*3)+1] = child2;
    }

    child3 = fork();
    if ( child3 == 0 )
    {
        if ( lameout )
        {
            lame_output_thread( &file_info );
        }

        if ( oggout )
        {
            oggenc_output_thread( &file_info );
        }

        exit(0);
    } else {
        children[(file_info.thread_num*3)+2] = child3;
    }

    waitpid( child1, NULL, 0 );
    waitpid( child2, NULL, 0 );
    waitpid( child3, NULL, 0 );

    children[(file_info.thread_num*3)] = -1;
    children[(file_info.thread_num*3)+1] = -1;
    children[(file_info.thread_num*3)+2] = -1;
    
#ifdef DEBUG
    fprintf( stderr, "Child #%d returned.\n", file_info.thread_num );
#endif
#endif

    processes[file_info.thread_num] = -1;
    
    sem_post( thread_count );

    if ( sem_close( thread_count ) == -1 )
    {
        snprintf( log_message, MAXLENGTH,
                  "sem_close() failed: %s\n", strerror( errno ) );
        print_log( log_message ); 
    }

    pthread_exit( &return_status );
}

/* Thread to read in the file and write it out to stdout.
 * We do this so that we can provide progress bars.
 */
void *file_write_thread( void *ptr )
{
    char log_message[MAXLENGTH];
    
    struct encode_data *file_info = (struct encode_data *) ptr;
    double last_progress = 0.0;
    double curr_progress = 0.0;
    mqd_t mque;
    struct bar_info_msg bar_data;
    
    int inputfd = 0;
    int fifo0fd = 0;

    char fifo0_name[MAXLENGTH];
    char fifo1_name[MAXLENGTH];

    struct stat stat_file_info;             /* File Info */
    int buffer_count = 0;
    int output_bytes = 0;
    char buffer[16384];

    mque = mq_open( AUDCONVERT_MQUE, O_WRONLY, NULL );
    if ( mque == -1 )
    {
        snprintf( log_message, MAXLENGTH,
                  "mq_open() failure: %s\n", strerror( errno ) );
        print_log( log_message ); 
    }

    sprintf( fifo0_name, "%s%d", AUDCONVERT_FIFO,
             file_info->thread_num * 2 );
    sprintf( fifo1_name, "%s%d", AUDCONVERT_FIFO, 
             (file_info->thread_num * 2) + 1 );

    fifo0fd = open( fifo0_name, O_WRONLY );
    if ( fifo0fd == -1 )
    {
        snprintf( log_message, MAXLENGTH,
                  "open of FIFO0 failed: %s\n", strerror( errno ) );
        print_log( log_message ); 
    }

    inputfd = open( file_info->file_name, O_RDONLY );
    if ( inputfd == -1 )
    {
        snprintf( log_message, MAXLENGTH,
                  "open of input file failed: %s\n", strerror( errno ) );
        print_log( log_message ); 
    }

    stat( file_info->file_name, &stat_file_info );
#ifdef DEBUG
    printf( "file size of %s = %d\n", file_info->file_name,
            (int) stat_file_info.st_size );
#endif

    while ( (output_bytes = read( inputfd, buffer, sizeof(char)*16384 ))
            != 0 )
    {
        if ( output_bytes != 
             write( fifo0fd, buffer, output_bytes ) )
        {
            snprintf( log_message, MAXLENGTH,
                      "ERROR: read bytes != write bytes\n" );
            print_log( log_message ); 
        }
    
        buffer_count += output_bytes;

        curr_progress = (double)buffer_count/(double)stat_file_info.st_size;

        if ( ((int) ((curr_progress - last_progress) * 100)) >= 1 )
        {
            last_progress = curr_progress;

            bar_data.mtype = PBARUPDATE;
            bar_data.bar_id = file_info->thread_num;
            sprintf( bar_data.bar_value, "%f", curr_progress);
                    
            if ( mq_send( mque, (char *) &bar_data,
                          sizeof( bar_data ), 0 ) != 0 )
            {
                snprintf( log_message, MAXLENGTH,
                          "mq_send() failed: %s\n", strerror( errno ) );
                print_log( log_message ); 
            } else {
#ifdef DEBUG
                printf( "SEND: pbar_status[%d] = %f\n",
                        file_info->thread_num,
                        atof(bar_data.bar_value) );
#endif
            }
        }
    }

    close( fifo0fd );
    close( inputfd );
    
    if ( mq_close( mque ) != 0 )
    {
        snprintf( log_message, MAXLENGTH,
                  "mq_close() failed: %s\n", strerror( errno ) );
        print_log( log_message ); 
    }

    return NULL;
}

/* Start the conversion process.  Wait for the end of the process. */
void *enter_dir_start( void *ptr )
{
    char log_message[MAXLENGTH];
    mqd_t mque;
    struct bar_info_msg bar_data;
    sem_t *thread_count;
    int i = 0;

    thread_count = sem_open( AUDCONVERT_SEM, 0 );
    if ( thread_count == SEM_FAILED )
    {
        snprintf( log_message, MAXLENGTH,
                  "semaphore create failed: %s\n", strerror( errno ) );
        print_log( log_message ); 
    }
    
    set_counting( 1 );
    files_processed = 0;

    enter_dir( input_dir_str, output_dir_str, thread_count );

    for ( i = 0; i < threads; i++ )
    {
        sem_wait( thread_count );
    }

    for ( i = 0; i < threads; i++ )
    {
        sem_post( thread_count );
    }
    
    /* Send DONE message. */
    mque = mq_open( AUDCONVERT_MQUE, O_WRONLY, NULL );
    if ( mque == -1 )
    {
        snprintf( log_message, MAXLENGTH,
                  "mq_open() failure: %s\n", strerror( errno ) );
        print_log( log_message ); 
    }

    bar_data.mtype = DONE;
    bar_data.bar_id = 0;
    sprintf( bar_data.bar_value, "%i", files_processed );
                    
    if ( mq_send( mque, (char *) &bar_data,
                  sizeof( bar_data ), 0 ) != 0 )
    {
        snprintf( log_message, MAXLENGTH,
                  "mq_send() failed: %s\n", strerror( errno ) );
        print_log( log_message ); 
    } else {
#ifdef DEBUG
        printf( "SEND: DONE = %d\n",
                atoi(bar_data.bar_value) );
#endif
    }

    if ( mq_close( mque ) != 0 )
    {
        snprintf( log_message, MAXLENGTH,
                  "mq_close() failed: %s\n", strerror( errno ) );
        print_log( log_message ); 
    }

    if ( sem_close( thread_count ) == -1 )
    {
        snprintf( log_message, MAXLENGTH,
                  "sem_close() failed: %s\n", strerror( errno ) );
        print_log( log_message ); 
    }

    return NULL;
}

/* This is where everything begins. */
int main( int argc, char **argv )
{
    int c = 0, i = 0;
    mqd_t mque;
    struct mq_attr attr;

    /* Process arguements. */
    while ( ( c = getopt( argc, argv, "bhfgi:o:t:v" ) ) != EOF ) {
        switch (c) {
            case 'b':
                justgo = 1;
                benchmark = 1;
                break;
            case 'h':
                if ( help_msg() == SUCCESS )
                    return EXIT_SUCCESS;
                else
                    return EXIT_FAILURE;
            case 'f':
                flatten = 1;
                break;
            case 'g':
                justgo = 1;
                break;
            case 'i':
                if ( strncpy( input_dir_str, optarg, MAXLENGTH ) == NULL )
                    fprintf( stderr,
                             "strncpy failed! Something is very wrong!\n" );
                break;
            case 'o':
                if ( strncpy( output_dir_str, optarg, MAXLENGTH ) == NULL )
                    fprintf( stderr,
                             "strncpy failed! Something is very wrong!\n" );
                break;
            case 't':
                threads = atoi( optarg );
                if ( threads < 1 )
                {
                    threads = 1;
                } else if ( threads > 8 ) {
                    threads = 8;
                }
                break;
            case 'v':
                printf( "\naudconvert version %s\n\n", VERSION );
                return EXIT_SUCCESS;
            case '?':
                return EXIT_FAILURE;
            default:
                fprintf( stderr, "Something has gone wrong!\n" );
                return EXIT_FAILURE;
        }
    }

    if ( (justgo == 1) && (strcmp( input_dir_str, "" ) == 0) &&
         (strcmp( output_dir_str, "" ) == 0) )
    {
        fprintf( stderr, "If you pass -g or -b you must also provide\n"
                         "an input and output directory using -i and -o.\n\n" );

        return EXIT_FAILURE;
    }

    for ( i = 0; i < MAX_THREADS; i++ )
    {
        processes[i] = -1;
    }

    for ( i = 0; i < (MAX_THREADS * 3); i++ )
    {
        children[ i ] = -1;
    }

    /* Make sure the message queue was deleted previously.
     * We'll ignore errors. */
    if ( mq_unlink( AUDCONVERT_MQUE ) != 0 )
    {
        /* Ignore Error. */
    }
    
    attr.mq_maxmsg = 10;
    attr.mq_msgsize = sizeof( struct bar_info_msg );
    
    /* Create the message queue then close it.
     * We'll unlink it when we shutdown. */
    mque = mq_open( AUDCONVERT_MQUE, O_CREAT, 0644, &attr, NULL );
    if ( mque == -1 )
    {
        perror( "mq_open() failure" );
    }

    if ( mq_close( mque ) != 0 )
    {
        perror( "mq_close() failed" );
    }

    /* Start the GUI. */
    start_gui( argc, argv, input_dir_str, output_dir_str );

    return EXIT_SUCCESS;
}
