#define _XOPEN_SOURCE

/* Includes */
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <signal.h>
#include <string.h>
#include <errno.h>
#include <mqueue.h>
#include <pthread.h>
#include <semaphore.h>
#include <sys/time.h>
#include <gtk/gtk.h>
#include "audconvert.h"
#include "gui.h"
/* End Includes */

/* non-ansi defines */
int snprintf(char *str, size_t size, const char *format, ...);

/* Debug Flag */
/* #define DEBUG */

/* Shared objects. */
static GtkWidget *window;
static GtkWidget *notebook;
static GtkWidget *pbars[MAX_THREADS];
static GtkWidget *sbars[MAX_THREADS];
static gint sbars_context_ids[MAX_THREADS];
static GtkWidget *input_text_entry;
static GtkWidget *output_text_entry;
static GtkWidget *elapsed_status;

static GtkWidget *progressbox;

static GtkWidget *flatbutton;

static GtkWidget *threadsentry;

static GtkWidget *ogginbutton;
static GtkWidget *mpg123inbutton;
static GtkWidget *flacinbutton;

static GtkWidget *lameoutbutton;
static GtkWidget *oggoutbutton;

static GtkWidget *bitrateentry;

static GtkTextBuffer *logbuffer = NULL;
static gint elapsed_context;
static int counting = 0;
static struct timeval start_time = { 0, 0 };

/* function defines */
gboolean delete_event( GtkWidget *widget, GdkEvent *event, gpointer data );
void about_app( void );

/* Menu Elements */
static GtkActionEntry entries[] = {
    { "FileMenu", NULL, "_File" },
    { "Exit", GTK_STOCK_QUIT, "E_xit", "<control>Q", "Exit the program",
      G_CALLBACK( delete_event ) },
    { "HelpMenu", NULL, "_Help" },
    { "About", GTK_STOCK_ABOUT, "_About", NULL, "About the program",
      about_app }
};

static const char *ui_description =
"<ui>"
"  <menubar name='MainMenu'>"
"    <menu action='FileMenu'>"
"      <menuitem action='Exit'/>"
"    </menu>"
"    <menu action='HelpMenu'>"
"      <menuitem action='About'/>"
"    </menu>"
"  </menubar>"
"</ui>";
/* End Menu Elements */

void about_app( void )
{
    char license[MAXLENGTH];

    strncpy( license, LICENSE1, MAXLENGTH );
    strncat( license, LICENSE2, MAXLENGTH );

    if ( strlen( license ) >= MAXLENGTH )
    {
        fprintf( stderr, "License is too big for alocated string.\n" );
    }
    
    gtk_show_about_dialog( GTK_WINDOW( window ), "name", APP_NAME,
                           "version", VERSION, "copyright", COPYRIGHT,
                           "license", license, "website", WEBSITE,
                           "website-label", WEBSITE_LABEL, NULL );
}

int cleanup( GtkWidget *widget, gint response, gpointer data )
{
    int i = 0;
    int threads = 0;
    char fifo0_name[MAXLENGTH];

    if ( response == GTK_RESPONSE_YES )
    {
        /* Kill children running. */
        kill_children();

        /* Cleanup semaphore. */
        if ( sem_unlink( AUDCONVERT_SEM ) == -1 )
        {
            /* Ignore Error. */
        }

        /* Cleanup message queue. */
        if ( mq_unlink( AUDCONVERT_MQUE ) != 0 )
        {
            perror( "mq_unlink() failed" );
        }

        /* Cleanup FIFOs. */
        threads = get_value( THREADS );
        for ( i = 0; i < threads; i++ )
        {
            sprintf( fifo0_name, "%s%d", AUDCONVERT_FIFO, i*2 );
            if ( unlink( fifo0_name ) != 0 )
            {
                if ( errno != ENOENT )
                {
                    perror( "Unlink of fifo failed" );
                }
            }

            sprintf( fifo0_name, "%s%d", AUDCONVERT_FIFO, (i*2)+1 );
            if ( unlink( fifo0_name ) != 0 )
            {
                if ( errno != ENOENT )
                {
                    perror( "Unlink of fifo failed" );
                }
            }
        }

        /* Clenaup GTK. */
        gtk_main_quit();
    }

    return TRUE;
}

gboolean delete_event( GtkWidget *widget, GdkEvent *event, gpointer data )
{
    GtkWidget *sureDialog;

    if ( counting == 1 )
    {
        sureDialog = gtk_message_dialog_new( GTK_WINDOW (window),
                                             GTK_DIALOG_DESTROY_WITH_PARENT,
                                             GTK_MESSAGE_QUESTION,
                                             GTK_BUTTONS_YES_NO,
                                             "Are you sure you wish to exit?" );

        g_signal_connect( G_OBJECT (sureDialog), "response",
                          G_CALLBACK (cleanup), NULL );
    
        gtk_dialog_run( GTK_DIALOG (sureDialog) );

        gtk_widget_destroy( sureDialog );
    } else {
        cleanup( NULL, GTK_RESPONSE_YES, NULL );
    }
    
    return TRUE;
}

int set_counting( int status )
{
    counting = status;

    return SUCCESS;
}

void print_log( char *log_message )
{
    GtkTextIter logiter;

    if ( logbuffer != NULL )
    {
        gtk_text_buffer_get_end_iter( logbuffer, &logiter );
        gtk_text_buffer_insert( logbuffer, &logiter, log_message, -1 );
    } else {
        fprintf( stderr, log_message );
    }
}

gboolean update_gui( gpointer ptr )
{
    char log_message[MAXLENGTH];
    int status = 0;
    struct bar_info_msg bar_data;
    mqd_t mque;
    struct timeval curr_time = { 0, 0 };
    int diff_seconds = 0;
    int diff_minutes = 0;
    int diff_hours = 0;
    char time_elapsed[MAXLENGTH];

    mque = mq_open( AUDCONVERT_MQUE, O_RDONLY|O_NONBLOCK, NULL );
    if ( mque == -1 )
    {
        perror( "mq_open() failure" );
    }

    status = mq_receive( mque, (char *) &bar_data, sizeof(bar_data), NULL );
    if ( status > 0 )
    {
        switch ( bar_data.mtype )
        {
            case PBARUPDATE:
#ifdef DEBUG
                printf( "RECV: Progress Bar[%d] = %f\n", bar_data.bar_id,
                                                     atof(bar_data.bar_value) );
#endif
                if ( atof(bar_data.bar_value) > 1.0 )
                {
                    snprintf( log_message, MAXLENGTH,
                              "ERROR: Progress greater than 1.0: %f\n",
                              atof(bar_data.bar_value) );
                    print_log( log_message );
                } else {
                    gtk_progress_bar_set_fraction( GTK_PROGRESS_BAR
                                                   (pbars[bar_data.bar_id]),
                                                   atof(bar_data.bar_value) );
                }
                break;
            case SBARUPDATE:
#ifdef DEBUG
                printf( "RECV: Status Bar[%d] = %s\n", bar_data.bar_id,
                                                       bar_data.bar_value );
#endif
                gtk_statusbar_push( GTK_STATUSBAR (sbars[bar_data.bar_id]),
                                    sbars_context_ids[bar_data.bar_id],
                                    bar_data.bar_value );

                snprintf( log_message, MAXLENGTH,
                          "Processing File: %s\n", bar_data.bar_value );
                print_log( log_message );

                break;
            case DONE:
                counting = 0;

                if ( gettimeofday( &curr_time, NULL ) != 0 )
                {
                    perror( "gettimeofday() returned failure" );
                } else {
                    if ( start_time.tv_sec == 0 )
                    {
                        memcpy( &start_time, &curr_time,
                                sizeof( struct timeval ) );
                    }

                    diff_seconds = (int) (curr_time.tv_sec - start_time.tv_sec);
                    diff_minutes = diff_seconds / 60;
                    diff_seconds = diff_seconds % 60;
                    diff_hours = diff_minutes / 60;
                    diff_minutes = diff_minutes % 60;

                    sprintf( time_elapsed, "Elapsed Time: %02d:%02d:%02d.%03d"
                             "  Process Complete.  Files Processed: %d",
                             diff_hours, diff_minutes, diff_seconds,
                             (int) (curr_time.tv_usec / 1000),
                             atoi(bar_data.bar_value) );
                    gtk_statusbar_push( GTK_STATUSBAR (elapsed_status),
                                        elapsed_context, time_elapsed );
                    start_time.tv_sec = 0;
                    start_time.tv_usec = 0;
                }

                if ( get_value( BENCHMARK ) == 1 )
                {
                    printf( "%s\n", time_elapsed );

                    sleep( 5 );
                    cleanup( NULL, GTK_RESPONSE_YES, NULL );
                }

                break;
            default:
                snprintf( log_message, MAXLENGTH,
                          "UNKNOWN MESSAGE!\n" );
                print_log( log_message );
                break;
        }
    } else {
        if ( errno != EAGAIN )
        {
            perror( "mq_receive() error" );
        }
    }

    if ( mq_close( mque ) != 0 )
    {
        perror( "mq_close() failed" );
    }

    if ( counting == 1 )
    {
        if ( gettimeofday( &curr_time, NULL ) != 0 )
        {
            perror( "gettimeofday() returned failure" );
        } else {
            if ( start_time.tv_sec == 0 )
            {
                memcpy( &start_time, &curr_time, sizeof( struct timeval ) );
            }

            diff_seconds = (int) (curr_time.tv_sec - start_time.tv_sec);
            diff_minutes = diff_seconds / 60;
            diff_seconds = diff_seconds % 60;
            diff_hours = diff_minutes / 60;
            diff_minutes = diff_minutes % 60;

            sprintf( time_elapsed, "Elapsed Time: %02d:%02d:%02d.%03d",
                     diff_hours, diff_minutes, diff_seconds,
                     (int) (curr_time.tv_usec / 1000) );
            gtk_statusbar_push( GTK_STATUSBAR (elapsed_status), elapsed_context,
                                time_elapsed );
        }
    }

    return TRUE;
}

int open_directory_dialog( GtkWidget *widget, gpointer data )
{
    char log_message[MAXLENGTH];
    GtkWidget *dialog;
    const gchar *buttontext = NULL;
    char *dirname = NULL;
    int *inputoutput;

    inputoutput = (int *) data;

    dialog = gtk_file_chooser_dialog_new( "Select Directory",
                                          GTK_WINDOW (window),
                                          GTK_FILE_CHOOSER_ACTION_SELECT_FOLDER,
                                          GTK_STOCK_CANCEL,
                                          GTK_RESPONSE_CANCEL,
                                          GTK_STOCK_OPEN,
                                          GTK_RESPONSE_ACCEPT,
                                          NULL);

    if ( gtk_dialog_run( GTK_DIALOG (dialog) ) == GTK_RESPONSE_ACCEPT )
    {
        /* dirname = gtk_file_chooser_get_current_folder( GTK_FILE_CHOOSER */
        dirname = gtk_file_chooser_get_filename( GTK_FILE_CHOOSER
                                                        (dialog) );

        buttontext = gtk_button_get_label( GTK_BUTTON (widget) );
        if ( *inputoutput  == INPUT )
        {
            gtk_entry_set_text( GTK_ENTRY (input_text_entry), dirname );
        } else if ( *inputoutput == OUTPUT ) {
            gtk_entry_set_text( GTK_ENTRY (output_text_entry), dirname );
        } else {
            snprintf( log_message, MAXLENGTH,
                      "Error: invalid button.\n" );
            print_log( log_message );
        }

        g_free( dirname );
    }

    gtk_widget_destroy( dialog );

    return( SUCCESS );
}

void start_convert( GtkWidget *widget, gpointer data )
{
    char log_message[MAXLENGTH];
    pthread_t threadid;
    char tmptxt[MAXLENGTH];
    int bitrate = 0;
    int threads = 0;
    sem_t *thread_count;
    char fifo0_name[MAXLENGTH];
    int i = 0;

    if ( counting == 1 )
    {
        return;
    }

    gtk_notebook_next_page( GTK_NOTEBOOK (notebook) );
        
    set_input_ouput_dir( gtk_entry_get_text( GTK_ENTRY (input_text_entry) ),
                         gtk_entry_get_text( GTK_ENTRY (output_text_entry) ) );
 
    set_value( FLATTEN, gtk_toggle_button_get_active( 
                        GTK_TOGGLE_BUTTON (flatbutton) ) );

    strcpy( tmptxt, gtk_entry_get_text( GTK_ENTRY (threadsentry) ) );
    threads = atoi( tmptxt );
    if ( threads < 1 )
    {
        snprintf( log_message, MAXLENGTH,
                  "Number of threads not valid.  Defaulting to 1.\n" );
        print_log( log_message );
        threads = 1;
    }

    if ( threads > MAX_THREADS )
    {
        snprintf( log_message, MAXLENGTH,
                  "Number of threads not valid.  Defaulting to %d.\n",
                  MAX_THREADS );
        print_log( log_message );
        threads = MAX_THREADS;
    }

    strcpy( tmptxt, gtk_entry_get_text( GTK_ENTRY (bitrateentry) ) );
    if ( atoi( tmptxt ) < 8 || atoi( tmptxt ) > 560 )
    {
        snprintf( log_message, MAXLENGTH,
                  "Output Bitrate not valid.  Defaulting to 192.\n" );
        print_log( log_message );
        bitrate = 192;
    } else {
        bitrate = atoi( tmptxt );
    }

    set_value( THREADS, threads );
    set_value( OUTPUTBITRATE, bitrate );
    set_value( OGGIN, gtk_toggle_button_get_active( 
                        GTK_TOGGLE_BUTTON (ogginbutton) ) );
    set_value( MP3IN, gtk_toggle_button_get_active( 
                        GTK_TOGGLE_BUTTON (mpg123inbutton) ) );
    set_value( FLACIN, gtk_toggle_button_get_active( 
                        GTK_TOGGLE_BUTTON (flacinbutton) ) );
    set_value( LAMEOUT, gtk_toggle_button_get_active( 
                        GTK_TOGGLE_BUTTON (lameoutbutton) ) );
    set_value( OGGOUT, gtk_toggle_button_get_active( 
                        GTK_TOGGLE_BUTTON (oggoutbutton) ) );

    /* Make sure the semaphore was deleted previously.  We'll ignore errors. */
    if ( sem_unlink( AUDCONVERT_SEM ) == -1 )
    {
        /* Ignore Error. */
    }

    /* Create the semaphore then close it. We'll unlink it when we shutdown. */
    thread_count = sem_open( AUDCONVERT_SEM, O_CREAT, 0644, threads );
    if ( thread_count == SEM_FAILED )
    {
        perror( "semaphore create failed" );
    }

    if ( sem_close( thread_count ) == -1 )
    {
        perror( "main sem_close() failed" );
    }

    for ( i = 0; i < threads; i++ )
    {
        sprintf( fifo0_name, "%s%d", AUDCONVERT_FIFO, i*2 );

#ifdef DEBUG
        fprintf( stderr, "Creating FIFO: %s\n", fifo0_name );
#endif

        if ( mkfifo( fifo0_name, 0777 ) == -1 )
        {
            if ( errno != EEXIST )
            {
                fprintf( stderr, "FIFO creation failed: %s: %s\n", fifo0_name,
                         strerror( errno ) );
            }
        }

        sprintf( fifo0_name, "%s%d", AUDCONVERT_FIFO, (i*2)+1 );

#ifdef DEBUG
        fprintf( stderr, "Creating FIFO: %s\n", fifo0_name );
#endif

        if ( mkfifo( fifo0_name, 0777 ) == -1 )
        {
            if ( errno != EEXIST )
            {
                fprintf( stderr, "FIFO creation failed: %s: %s\n", fifo0_name,
                         strerror( errno ) );
            }
        }
    }

    /* Progress Tab */
    for ( i = 0; i < threads; i++ )
    {
        if ( pbars[i] == NULL )
        {
            pbars[i] = gtk_progress_bar_new();
            gtk_box_pack_start( GTK_BOX(progressbox), pbars[i],
                                FALSE, FALSE, 0 );

            sbars[i] = gtk_statusbar_new();
            gtk_statusbar_set_has_resize_grip( GTK_STATUSBAR (sbars[i]),
                                               FALSE );
            gtk_box_pack_start( GTK_BOX(progressbox), sbars[i],
                                FALSE, FALSE, 0 );

            sbars_context_ids[i] = gtk_statusbar_get_context_id(
                                   GTK_STATUSBAR (sbars[i]), "DEFAULT" );
        }
        gtk_widget_show( pbars[i] );
        gtk_widget_show( sbars[i] );
    }

    for ( ; i < MAX_THREADS; i++ )
    {
        if ( pbars[i] != NULL )
        {
            gtk_widget_hide( pbars[i] );
            gtk_widget_hide( sbars[i] );
        }
    }
    /* End Progress Tab */
    
    pthread_create( &threadid, NULL, enter_dir_start, NULL );
    pthread_detach( threadid );
}

int start_gui( int argc, char **argv,
               char *input_dir_str, char *output_dir_str )
{
    GtkWidget *inoutlabel;
    GtkWidget *progresslabel;
    GtkWidget *loglabel;
    
    GtkWidget *blanklabel01;
    GtkWidget *blanklabel02;
    GtkWidget *blanklabel03;
    GtkWidget *blanklabel04;
    GtkWidget *blanklabel05;
    
    GtkWidget *mainbox;
    GtkWidget *logbox;
    
    GtkWidget *inouttable;
    GtkWidget *inputbox;
    GtkWidget *inputframe;
    GtkWidget *outputbox;
    GtkWidget *outputframe;

    GtkWidget *inputlabel;
    GtkWidget *outputlabel;
    GtkWidget *inputbutton;
    GtkWidget *outputbutton;
    GtkWidget *threadslabel;
    GtkWidget *gobutton;
    GtkWidget *inavaillabel;
    GtkWidget *outformatlabel;

    GtkWidget *inavailalign;
    GtkWidget *outformatalign;

    GtkWidget *bitratealign;
    GtkWidget *bitratebox;
    GtkWidget *bitratelabel;
    GtkWidget *logwindow;
    GtkWidget *logview;

    /* Menu Items */
    GtkWidget *menubar;
    GtkActionGroup *action_group;
    GtkUIManager *ui_manager;
    GtkAccelGroup *accel_group;
    GError *error;
    /* end Menu Items */

    int *inid;
    int *outid;

    char tmptxt[MAXLENGTH];

    int i = 0;

    inid = malloc( sizeof( int ) );
    outid = malloc( sizeof( int ) );

    gtk_init( &argc, &argv );

    blanklabel01 = gtk_label_new( "" );
    blanklabel02 = gtk_label_new( "" );
    blanklabel03 = gtk_label_new( "" );
    blanklabel04 = gtk_label_new( "" );
    blanklabel05 = gtk_label_new( "" );

    /* Main Window */
    window = gtk_window_new( GTK_WINDOW_TOPLEVEL );

    gtk_window_set_title( GTK_WINDOW (window), TITLE_BAR );
    gtk_widget_set_size_request( GTK_WIDGET (window), 600, 600 );

    g_signal_connect( G_OBJECT (window), "delete_event",
                      G_CALLBACK (delete_event), NULL );

    gtk_container_set_border_width( GTK_CONTAINER (window), 1 );
    /* End Main Window */

    /* Main Packing Containers*/
    /* Main Box */
    mainbox = gtk_vbox_new( FALSE, 10 );
    gtk_container_add( GTK_CONTAINER (window), mainbox );

    /* Main Notebook */
    notebook = gtk_notebook_new();
    gtk_notebook_set_tab_pos( GTK_NOTEBOOK (notebook), GTK_POS_TOP );

    inoutlabel = gtk_label_new( "Input/Output" );
    inouttable = gtk_table_new( 9, 4, FALSE );
    gtk_notebook_append_page( GTK_NOTEBOOK (notebook), inouttable,
                              inoutlabel);

    progresslabel = gtk_label_new( "Progress" );
    progressbox = gtk_vbox_new( FALSE, 5 );
    gtk_notebook_append_page( GTK_NOTEBOOK (notebook), progressbox,
                              progresslabel );

    loglabel = gtk_label_new( "Log" );
    logbox = gtk_vbox_new( FALSE, 5 );
    gtk_notebook_append_page( GTK_NOTEBOOK (notebook), logbox, loglabel );
    /* End Main Packing Container */

    /* Menu */
    action_group = gtk_action_group_new( "MenuActions" );
    gtk_action_group_add_actions( action_group, entries,
                                  G_N_ELEMENTS (entries), window );

    ui_manager = gtk_ui_manager_new();
    gtk_ui_manager_insert_action_group( ui_manager, action_group, 0 );

    accel_group = gtk_ui_manager_get_accel_group( ui_manager );
    gtk_window_add_accel_group( GTK_WINDOW (window), accel_group );

    error = NULL;
    if ( !gtk_ui_manager_add_ui_from_string( ui_manager, ui_description,
                                             -1, &error ) )
    {
        g_message ("building menus failed: %s", error->message);
        g_error_free (error);
        exit (EXIT_FAILURE);
    }

    menubar = gtk_ui_manager_get_widget( ui_manager, "/MainMenu" );
    gtk_box_pack_start( GTK_BOX (mainbox), menubar, FALSE, FALSE, 0 );
    /* End Menu */

    /* Now that we have a menu, pack the notebook. */
    gtk_box_pack_start( GTK_BOX (mainbox), notebook, TRUE, TRUE, 0 );

    /* Input/Output Tab */
    inputlabel = gtk_label_new( "Input Directory:" );
    input_text_entry = gtk_entry_new();
    if ( input_dir_str != NULL )
    {
        gtk_entry_set_text( GTK_ENTRY (input_text_entry), input_dir_str );
    }
    inputbutton = gtk_button_new_from_stock( GTK_STOCK_OPEN );
    *inid = INPUT;
    g_signal_connect ( G_OBJECT (inputbutton), "clicked",
                       G_CALLBACK (open_directory_dialog), inid );
    gtk_table_attach( GTK_TABLE (inouttable), inputlabel,
                      0, 1, 0, 1, GTK_SHRINK, GTK_SHRINK, 2, 2 );
    gtk_table_attach( GTK_TABLE (inouttable), input_text_entry,
                      1, 3, 0, 1, GTK_EXPAND|GTK_FILL, GTK_SHRINK, 2, 2 );
    gtk_table_attach( GTK_TABLE (inouttable), inputbutton,
                      3, 4, 0, 1, GTK_SHRINK, GTK_SHRINK, 2, 2 );

    outputlabel = gtk_label_new( "Output Directory:" );
    output_text_entry = gtk_entry_new();
    if ( output_dir_str != NULL )
    {
        gtk_entry_set_text( GTK_ENTRY (output_text_entry), output_dir_str );
    }
    outputbutton = gtk_button_new_from_stock( GTK_STOCK_OPEN );
    *outid = OUTPUT;
    g_signal_connect ( G_OBJECT (outputbutton), "clicked",
                       G_CALLBACK (open_directory_dialog), outid );
    gtk_table_attach( GTK_TABLE (inouttable), outputlabel,
                      0, 1, 1, 2, GTK_SHRINK, GTK_SHRINK, 2, 2 );
    gtk_table_attach( GTK_TABLE (inouttable), output_text_entry,
                      1, 3, 1, 2, GTK_EXPAND|GTK_FILL, GTK_SHRINK, 2, 2 );
    gtk_table_attach( GTK_TABLE (inouttable), outputbutton,
                      3, 4, 1, 2, GTK_SHRINK, GTK_SHRINK, 2, 2 );

    flatbutton = gtk_check_button_new_with_mnemonic(
                     "Flatten output directory structure." );
    gtk_toggle_button_set_active( GTK_TOGGLE_BUTTON (flatbutton),
                                  get_value( FLATTEN ) );
    gtk_table_attach( GTK_TABLE (inouttable), flatbutton,
                      1, 3, 2, 3, GTK_FILL, GTK_SHRINK, 2, 2 );

    /* Seperator. */
    gtk_table_attach( GTK_TABLE (inouttable), blanklabel03,
                      0, 4, 3, 4, GTK_FILL, GTK_SHRINK, 2, 2 );

    /* Thread Entry. */
    snprintf( tmptxt, MAXLENGTH, "Number of Threads (MAX: %d):",
              MAX_THREADS );
    threadslabel = gtk_label_new( tmptxt );
    threadsentry = gtk_entry_new();
    gtk_entry_set_max_length( GTK_ENTRY (threadsentry), 3 );
    gtk_entry_set_width_chars( GTK_ENTRY (threadsentry), 4 );
    snprintf( tmptxt, 4, "%d", get_value( THREADS ) );
    gtk_entry_set_text( GTK_ENTRY (threadsentry), tmptxt );
    gtk_table_attach( GTK_TABLE (inouttable), threadslabel,
                      0, 2, 4, 5, GTK_FILL, GTK_SHRINK, 2, 2 );
    gtk_table_attach( GTK_TABLE (inouttable), threadsentry,
                      2, 3, 4, 5, GTK_FILL, GTK_SHRINK, 2, 2 );

    /* Seperator. */
    gtk_table_attach( GTK_TABLE (inouttable), blanklabel04,
                      0, 4, 5, 6, GTK_FILL, GTK_SHRINK, 2, 2 );

    /* Option Boxes. */
    inputbox = gtk_vbox_new( FALSE, 2 );
    inputframe = gtk_frame_new( "Input Options" );
    outputbox = gtk_vbox_new( FALSE, 2 );
    outputframe = gtk_frame_new( "Output Options" );
    gtk_container_add( GTK_CONTAINER (inputframe), inputbox );
    gtk_container_add( GTK_CONTAINER (outputframe), outputbox );
    gtk_table_attach( GTK_TABLE (inouttable), inputframe,
                      0, 2, 6, 7, GTK_EXPAND|GTK_FILL, GTK_FILL, 2, 2 );
    gtk_table_attach( GTK_TABLE (inouttable), outputframe,
                      2, 4, 6, 7, GTK_EXPAND|GTK_FILL, GTK_FILL, 2, 2 );

    /* Seperator. */
    gtk_table_attach( GTK_TABLE (inouttable), blanklabel05,
                      0, 4, 7, 8, GTK_FILL, GTK_SHRINK, 2, 2 );

    /* Go button. */
    gobutton = gtk_button_new_with_label( "Go" );
    g_signal_connect ( G_OBJECT (gobutton), "clicked",
                       G_CALLBACK (start_convert), NULL);
    gtk_table_attach( GTK_TABLE (inouttable), gobutton,
                      1, 3, 8, 9, GTK_FILL, GTK_SHRINK, 2, 2 );
    
    inavailalign = gtk_alignment_new( 0, 0, 0, 0 );
    inavaillabel = gtk_label_new( "Process Input Formats:" );
    gtk_container_add( GTK_CONTAINER (inavailalign), inavaillabel );
    gtk_box_pack_start_defaults( GTK_BOX (inputbox), inavailalign );
    outformatalign = gtk_alignment_new( 0, 0, 0, 0 );
    outformatlabel = gtk_label_new( "Output Format:" );
    gtk_container_add( GTK_CONTAINER (outformatalign), outformatlabel );
    gtk_box_pack_start_defaults( GTK_BOX (outputbox), outformatalign );

    ogginbutton = gtk_check_button_new_with_mnemonic(
                     "Ogg Vorbis (oggdec)" );
    gtk_toggle_button_set_active( GTK_TOGGLE_BUTTON (ogginbutton),
                                  get_value( OGGIN ) );
    mpg123inbutton = gtk_check_button_new_with_mnemonic(
                     "MP3 (mpg123)" );
    gtk_toggle_button_set_active( GTK_TOGGLE_BUTTON (mpg123inbutton),
                                  get_value( MP3IN ) );
    flacinbutton = gtk_check_button_new_with_mnemonic(
                     "FLAC (flac)" );
    gtk_toggle_button_set_active( GTK_TOGGLE_BUTTON (flacinbutton),
                                  get_value( FLACIN ) );
    gtk_box_pack_start_defaults( GTK_BOX (inputbox), ogginbutton );
    gtk_box_pack_start_defaults( GTK_BOX (inputbox), mpg123inbutton );
    gtk_box_pack_start_defaults( GTK_BOX (inputbox), flacinbutton );

    gtk_box_pack_start_defaults( GTK_BOX (inputbox), blanklabel01 );

    lameoutbutton = gtk_radio_button_new_with_mnemonic( NULL,
                                                        "MP3 (lame)" );
    oggoutbutton = gtk_radio_button_new_with_mnemonic_from_widget( 
                        GTK_RADIO_BUTTON (lameoutbutton),
                        "Ogg Vorbis (oggenc)" );
    gtk_box_pack_start_defaults( GTK_BOX (outputbox), lameoutbutton );
    gtk_box_pack_start_defaults( GTK_BOX (outputbox), oggoutbutton );

    gtk_box_pack_start_defaults( GTK_BOX (outputbox), blanklabel02 );

    bitratebox = gtk_hbox_new( FALSE, 2 );
    bitratelabel = gtk_label_new( "Output Bitrate (kbps): " );
    bitrateentry = gtk_entry_new();
    gtk_entry_set_max_length( GTK_ENTRY (bitrateentry), 3 );
    gtk_entry_set_width_chars( GTK_ENTRY (bitrateentry), 4 );
    snprintf( tmptxt, 4, "%d", get_value( OUTPUTBITRATE ) );
    gtk_entry_set_text( GTK_ENTRY (bitrateentry), tmptxt );
    gtk_box_pack_start_defaults( GTK_BOX (bitratebox), bitratelabel );
    gtk_box_pack_start_defaults( GTK_BOX (bitratebox), bitrateentry );
    bitratealign = gtk_alignment_new( 0, 0, 0, 0 );
    gtk_container_add( GTK_CONTAINER (bitratealign), bitratebox );
    gtk_box_pack_start_defaults( GTK_BOX (outputbox), bitratealign );
    /* End Input/Output Tab */

    /* Setup/NULL pbars, sbars, and sbars_context_ids. */
    for ( i = 0; i < MAX_THREADS; i++ )
    {
        pbars[i] = NULL;
        sbars[i] = NULL;
        sbars_context_ids[i] = 0;
    }

    /* Add timer to update GUI. */
    g_timeout_add( 100, update_gui, NULL );

    /* Log Tab */
    logwindow = gtk_scrolled_window_new( NULL, NULL );

    logview = gtk_text_view_new();
    gtk_text_view_set_editable( GTK_TEXT_VIEW (logview), FALSE );
    gtk_text_view_set_cursor_visible( GTK_TEXT_VIEW (logview), FALSE );
    logbuffer = gtk_text_view_get_buffer( GTK_TEXT_VIEW (logview) );

    gtk_scrolled_window_add_with_viewport( GTK_SCROLLED_WINDOW (logwindow),
                                           logview );

    gtk_box_pack_start( GTK_BOX(logbox), logwindow, TRUE, TRUE, 0 );
    /* End Log Tab */

    /* Overall Status Bar */
    elapsed_status = gtk_statusbar_new();
    gtk_box_pack_end( GTK_BOX(mainbox), elapsed_status, FALSE, FALSE, 0 );
    elapsed_context = gtk_statusbar_get_context_id(
                                GTK_STATUSBAR (elapsed_status),
                                "DEFAULT" );
    /* End Overall Status Bar */

    gtk_widget_show_all( window );

    if ( get_value( JUSTGO ) == 1 )
    {
        gtk_init_add( (gboolean (*)(gpointer))start_convert, NULL );
    }
    
    gtk_main ();

    return SUCCESS;
}
